"""Functions for applying dynamic real-valued shifts to numpy arrays using Lagrange interpolation

Use make_dynamic_shift_lagrange_numpy to create a Lagrange interpolator for numpy arrays.
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Final

import numpy as np
from typing_extensions import assert_never

from lisainstrument.fir_filters_numpy import NumpyArray1D
from lisainstrument.regular_interpolators import (
    RegularInterpolator,
    make_regular_interpolator_lagrange,
    make_regular_interpolator_linear,
)


class ShiftBC(Enum):
    """Enum for various methods of handling boundaries in dynamic shift

    ZEROPAD: Assume all data outside given range is zero.
    FLAT: Assume all data outside given range is equal to value on nearest boundary
    EXCEPTION: raise exception if data outside range is needed
    """

    EXCEPTION = 1
    ZEROPAD = 2
    FLAT = 3


@dataclass(frozen=True)
class DynShiftCfg:
    """Config class for dynamic shift interpolation

    Attributes:
        min_delay: Assume that any shift > -max_delay
        max_delay: Assume that any shift < -min_delay
        left_bound: Treatment of left boundary
        right_bound: Treatment of right boundary
    """

    min_delay: float
    max_delay: float
    left_bound: ShiftBC
    right_bound: ShiftBC

    @property
    def min_delay_int(self) -> int:
        """Minimum delay in integer index space"""
        return int(np.floor(self.min_delay))

    @property
    def max_delay_int(self) -> int:
        """Maximum delay in integer index space"""
        return int(np.ceil(self.max_delay))

    def __post__init__(self):
        if self.min_delay > self.max_delay:
            msg = f"Invalid delay range for dynamic shift ({self.min_delay}, {self.max_delay})"
            raise ValueError(msg)


class DynamicShiftNumpy:
    """Interpolate to locations given by a non-const shift.

    This allows to interpolate samples in a numpy array to locations specified
    by a shift given by another numpy array of same size. The shift is specified in
    units of the array index, i.e. there is no separate coordinate array.
    A positive shift refers to values right of a given sample, negative shifts
    to values on the left.

    The boundary treatment can be specified for each boundary in terms of
    ShiftBC enums.

    The interpolation method is not fixed but provided via an interpolator
    instance implementing the RegularInterpMethod protocol.

    For technical reasons, the shift values have to be within some bound that
    has to be provided.
    """

    def __init__(
        self,
        cfg: DynShiftCfg,
        interp: RegularInterpolator,
    ):
        """Set up interpolator.

        Arguments:
            cfg: limits of delay and boundary treatment
            interp: interpolation method
        """
        self._cfg: Final = cfg
        self._interp_np: Final = interp

    @property
    def margin_left(self) -> int:
        """Left margin size.

        If positive, specifies how many samples on the left have to be added by boundary conditions.
        If negative, specifies how many samples on the left are unused.
        """
        return self._interp_np.margin_left + self._cfg.max_delay_int

    @property
    def margin_right(self) -> int:
        """Right margin size.

        If positive, specifies how many samples on the right have to be added by boundary conditions.
        If negative, specifies how many samples on the right are unused.
        """
        return self._interp_np.margin_right - self._cfg.min_delay_int

    def __call__(self, samples: np.ndarray, shift: np.ndarray) -> NumpyArray1D:
        """Apply shift given by numpy array to samples in another numpy array.

        The shift and sample arrays need to have the same size, and each shift provides
        the interpolation location relative to the sample with the same index.
        Shifts are floating point values. A shift of +1 refers to the sample on the right,
        -1 the sample on the left, etc. All arrays have to be 1D.


        Arguments:
            samples: 1D numpy array with data samples
            shift: 1D numpy array with shifts

        Returns:
            Numpy array with interpolated samples
        """
        out_size = len(shift)

        npad_left = 0
        if self.margin_left > 0:
            match self._cfg.left_bound:
                case ShiftBC.ZEROPAD:
                    npad_left = self.margin_left
                    samples = np.concatenate([np.zeros(npad_left), samples])
                case ShiftBC.FLAT:
                    npad_left = self.margin_left
                    samples = np.concatenate([np.ones(npad_left) * samples[0], samples])
                case ShiftBC.EXCEPTION:
                    msg = (
                        f"DynamicShiftNumpy: left edge handling {self._cfg.left_bound.name} not "
                        f"possible for given max delay {self._cfg.max_delay}."
                    )
                    raise RuntimeError(msg)
                case _ as unreachable:
                    assert_never(unreachable)

        if self.margin_right > 0:
            match self._cfg.right_bound:
                case ShiftBC.ZEROPAD:
                    samples = np.concatenate([samples, np.zeros(self.margin_right)])
                case ShiftBC.FLAT:
                    samples = np.concatenate(
                        [samples, np.ones(self.margin_right) * samples[-1]]
                    )
                case ShiftBC.EXCEPTION:
                    msg = (
                        f"DynamicShiftNumpy: right edge handling {self._cfg.right_bound.name} not "
                        f"possible for given min delay {self._cfg.min_delay=}."
                    )
                    raise RuntimeError(msg)
                case _ as unreachable:
                    assert_never(unreachable)

        pos = 0
        n_size = npad_left + out_size + self.margin_right
        n_first = pos + npad_left - self.margin_left
        samples_needed = samples[n_first : n_first + n_size]
        return self._interp_np.apply_shift(samples_needed, shift, self.margin_left)


def make_dynamic_shift_lagrange_numpy(
    order: int,
    min_delay: float,
    max_delay: float,
    left_bound: ShiftBC,
    right_bound: ShiftBC,
) -> DynamicShiftNumpy:
    """Set up DynamicShiftNumpy instance with Lagrange interpolation method.

    Arguments:
        order: Order of the Lagrange polynomials
        min_delay: Assume that any shift > -max_delay
        max_delay: Assume that any shift < -min_delay
        left_bound: Treatment of left boundary
        right_bound: Treatment of right boundary

    Returns:
        Interpolation function of type DynamicShiftNumpy
    """
    interp = make_regular_interpolator_lagrange(order)
    cfg = DynShiftCfg(min_delay, max_delay, left_bound, right_bound)
    return DynamicShiftNumpy(cfg, interp)


def make_dynamic_shift_linear_numpy(
    min_delay: float,
    max_delay: float,
    left_bound: ShiftBC,
    right_bound: ShiftBC,
) -> DynamicShiftNumpy:
    """Set up DynamicShiftNumpy instance with linear interpolation method.

    Arguments:
        min_delay: Assume that any shift > -max_delay
        max_delay: Assume that any shift < -min_delay
        left_bound: Treatment of left boundary
        right_bound: Treatment of right boundary

    Returns:
        Interpolation function of type DynamicShiftNumpy
    """
    interp = make_regular_interpolator_linear()
    cfg = DynShiftCfg(min_delay, max_delay, left_bound, right_bound)
    return DynamicShiftNumpy(cfg, interp)
